@@ -7,32 +7,35 @@ if six.PY3:
     from builtins import input
     from builtins import object
 
-#import all of this version information
-__version__ = '0.2.8'
-__author__ = 'dave@springserve.com'
-__license__ = 'Apache 2.0'
-__copyright__ = 'Copyright 2016 Springserve'
-__title__ = 'springserve'
+
+__version__ = '0.8.6' #TODO: This is duplicated in the build.  Need to figure how to set this once 
 
 import sys as _sys
 import json as _json
 import getpass
 
+from requests_toolbelt import MultipartEncoder
+
 _msg = None
 
 try:
     from link import lnk as _lnk
+    from link.wrappers import SpringServeAPI, SpringAuth
     _msg = _lnk.msg
-except: 
+except:
     print("problem loading link, this is ok on the install")
 
+from ._decorators import raw_response_retry
+
 _API = None
+_TOKEN_OVERRIDE = None
 _ACCOUNT = None
-_DEFAULT_BASE_URL = "https://video.springserve.com/api/v0"
+_DEFAULT_BASE_URL = "https://console.springserve.com/api/v0"
+_CONFIG_OVERRIDE = None
 
 def setup_config():
     """
-    This is used the first time you run it to set up your configuration if 
+    This is used the first time you run it to set up your configuration if
     you want to do that over the command prompt
     """
     current_file = _lnk.config_file()
@@ -62,70 +65,110 @@ def setup_config():
     if not confirm or confirm.upper() != 'Y':
         print("thanks")
         return
-   
+
     print("writing config to: {}".format(current_file))
     with open(current_file, 'w') as f:
         f.write(_json.dumps(current_config, indent=4))
 
-    print("done: refreshing config") 
+    print("done: refreshing config")
     _lnk.fresh()
 
+
+class SpringServeAPITokenOverride(SpringServeAPI):
+    """
+    Used to inject a token instead of having the SDK do auth
+    """
+
+    def __init__(self, base_url=None, user=None, password=None, token=None):
+        self._token = token
+        super(SpringServeAPI, self).__init__(base_url=base_url,
+                                             user=user,
+                                             password=password)
+
+    def authenticate(self):
+        self._wrapped.auth = SpringAuth(self._token)
+
+
 def API(reauth=False):
     """
     Get the raw API object.  This is rarely used directly by a client of this
     library, but it used as an internal function
     """
-    global _API, _ACCOUNT
+    global _API, _ACCOUNT, _CONFIG_OVERRIDE, _TOKEN_OVERRIDE
 
     if _API is None or reauth:
         _msg.debug("authenticating to springserve")
         try:
             if _ACCOUNT:
                 _API = _lnk("springserve.{}".format(_ACCOUNT))
+            elif _CONFIG_OVERRIDE:
+                _API = SpringServeAPI(**_CONFIG_OVERRIDE)
+
+            elif _TOKEN_OVERRIDE:
+                _API = SpringServeAPITokenOverride(**_TOKEN_OVERRIDE)
             else:
                 try:
-                    _API = _lnk("springserve.{}".format("__default__")) 
+                    _API = _lnk("springserve.{}".format("__default__"))
                 except:
-                    #this is to keep backwards compatiblity
+                    # this is to keep backwards compatiblity
                     _API = _lnk.springserve
-                   
-
+            _API.headers.update({'springserve-sdk': __version__})
         except Exception as e:
             raise Exception("""Error authenticating: check your link.config to
                             make sure your username, password and url are
-                            correct""" )
+                            correct: {}""".format(e))
     return _API
 
+
 def switch_account(account_name="__default__"):
     global _ACCOUNT
     _ACCOUNT = account_name
-    API(True) 
+    API(True)
+
+def set_credentials(user, password, base_url='https://console.springserve.com/api/v0'):
+    global _CONFIG_OVERRIDE
+
+    _CONFIG_OVERRIDE = {'user': user, 'password': password, 'base_url': base_url} 
+    API(True)
+
+def set_token(token, base_url='https://console.springserve.com/api/v0'):
+    global _TOKEN_OVERRIDE
+
+    _TOKEN_OVERRIDE = {'token': token, 'base_url': base_url}
+    API(True)
+
 
 class _TabComplete(object):
     """
-    this class exists to make any other class 
-    have a tab completion function that is already 
-    hooked into ipython 
+    this class exists to make any other class
+    have a tab completion function that is already
+    hooked into ipython
     """
     def _tab_completions(self):
         return []
 
+    def __dir__(self):
+        return super(_TabComplete, self).__dir__() + self._tab_completions()
+
 
 class _VDAPIResponse(_TabComplete):
 
-    def __init__(self, service, api_response_data, path_params, query_params, ok):
+    def __init__(self, service, api_response_data, path_params, query_params,
+                 ok, payload='', injected_account_id=None):
         super(_VDAPIResponse, self).__init__()
         self._service = service
         self._raw_response = api_response_data
         self._path_params = path_params
         self._query_params = query_params or {}
         self._ok = ok
+        self._payload = payload
+        self._injected_account_id = injected_account_id
 
     @property
     def ok(self):
         """
-        Tells you if the api response was "ok" 
-        
+        Tells you if the api response was "ok"
+
         meaning that it responded with a 200.  If there was an error, this will
         return false
 
@@ -157,7 +200,7 @@ class _VDAPIResponse(_TabComplete):
         try:
             return self.__getattribute__(key)
         except AttributeError as e:
-            #makes unpickling work?
+            # makes unpickling work?
             if key.startswith("__"):
                 raise e
             return self._raw_response[key]
@@ -172,14 +215,17 @@ class _VDAPIResponse(_TabComplete):
 
 class _VDAPISingleResponse(_VDAPIResponse):
 
-    def __init__(self, service, api_response_data, path_params, query_params, ok):
+    def __init__(self, service, api_response_data, path_params, query_params,
+                 ok, payload='', injected_account_id=None):
         self._dirty = {}
         super(_VDAPISingleResponse, self).__init__(service, api_response_data,
-                                                   path_params, query_params, ok)
-    
+                                                   path_params, query_params,
+                                                   ok, payload,
+                                                   injected_account_id)
+
     def set_dirty(self, field):
         """
-        you need this for nested fields that you have changed 
+        you need this for nested fields that you have changed
         but didn't actually set
         """
         self._dirty[field] = self._raw_response[field]
@@ -195,23 +241,30 @@ class _VDAPISingleResponse(_VDAPIResponse):
 
         Returns:
 
-            An API response object 
+            An API response object
         """
-        #if they have dirty fields only send those
+        # if they have dirty fields only send those
 
         payload = self.raw
 
         if dirty_only:
             payload = self._dirty
 
-        return self._service.put(self.id, payload, account_id =
-                                 self.account_id, **kwargs)
-    
+        try:
+            account_id = self.account_id
+        except Exception as e:
+            if self._injected_account_id:
+                account_id = self._injected_account_id
+            else:
+                raise e
+
+        return self._service.put(self.id, payload, account_id=account_id, **kwargs)
+
     def duplicate(self, **kwargs):
 
         payload = self.raw.copy()
         payload.update(kwargs)
-        return self._service.new(payload, account_id = self.account_id)
+        return self._service.new(payload, account_id=self.account_id)
 
     def __setattr__(self, attr, value):
         """
@@ -236,15 +289,17 @@ class _VDAPISingleResponse(_VDAPIResponse):
 class _VDAPIMultiResponse(_VDAPIResponse):
 
     def __init__(self, service, api_response_data, path_params, query_params,
-                 response_object, ok):
+                 response_object, ok, payload='', injected_account_id=None):
 
         super(_VDAPIMultiResponse, self).__init__(service, api_response_data,
-                                                  path_params, query_params, ok)
+                                                  path_params, query_params, ok, payload)
+        self._payload = payload
         self._object_cache = []
         self._current_page = 1
         self._all_pages_gotten = False
+        self._injected_account_id = injected_account_id
         self.response_object = response_object
-        #build out the initial set of objects
+        # build out the initial set of objects
         self._build_cache(self.raw)
 
     def _build_cache(self, objects):
@@ -252,32 +307,36 @@ class _VDAPIMultiResponse(_VDAPIResponse):
                                    objects])
 
     def _is_last_page(self, resp):
-        #this means we 
         return (not resp or not resp.json)
 
     def _get_next_page(self):
-        
+
         if self._all_pages_gotten:
-            return 
+            return
 
         params = self._query_params.copy()
         params['page'] = self._current_page+1
         resp = self._service.get_raw(self._path_params, **params)
 
-        # this means we are donesky, we don't know 
+        # this means we are donesky, we don't know
         # how many items there will be, only that we hit the last page
         if self._is_last_page(resp):
             self._all_pages_gotten = True
             return
 
         self._build_cache(resp.json)
-        self._current_page += 1 
+        self._current_page += 1
 
     def _build_response_object(self, data):
         return self.response_object(self._service, data,
                                     self._path_params,
-                                    self._query_params, True)
+                                    self._query_params,
+                                    True,
+                                    payload='',
+                                    injected_account_id=self._injected_account_id)
+
     def __getitem__(self, key):
+
         if not isinstance(key, int):
             raise Exception("Must be an index ")
         if key >= len(self._object_cache):
@@ -289,11 +348,11 @@ class _VDAPIMultiResponse(_VDAPIResponse):
 
     def __iter__(self):
         """
-        this will automatically take care of pagination for us. 
+        this will automatically take care of pagination for us.
         """
         idx = 0
         while True:
-            # not sure I love this method, but it's the best 
+            # not sure I love this method, but it's the best
             # one I can think of right now
             try:
                 yield self[idx]
@@ -314,22 +373,30 @@ def _format_url(endpoint, path_param):
 
     return _url
 
+
 def _format_params(params):
-    
+
     _params = {}
 
     for key, value in params.items():
         if isinstance(value, list):
-            #make sure any list has the [] on it 
+            # make sure any list has the [] on it
             key = "{}[]".format(key.lstrip("[]"))
         _params[key] = value
 
     return _params
 
 
+class _VDDuplicateableResponse(_VDAPISingleResponse):
+
+    def duplicate(self, **kwargs):
+        return self._service.get("{}/duplicate".format(self.id), account_id=self.account_id)
+
+
 class VDAuthError(Exception):
     pass
 
+
 class _VDAPIService(object):
 
     __API__ = None
@@ -337,13 +404,13 @@ class _VDAPIService(object):
     __RESPONSES_OBJECT__ = _VDAPIMultiResponse
 
     def __init__(self):
-        pass 
+        self.account_id = None
 
     @property
     def endpoint(self):
         """
-        The api endpoint that is used for this service.  For example:: 
-            
+        The api endpoint that is used for this service.  For example::
+
             In [1]: import springserve
 
             In [2]: springserve.supply_tags.endpoint
@@ -352,26 +419,31 @@ class _VDAPIService(object):
         """
         return "/" + self.__API__
 
-    def build_response(self, api_response, path_params, query_params):
+    def build_response(self, api_response, path_params, query_params, payload=''):
         is_ok = api_response.ok
 
         if not is_ok and api_response.status_code == 401:
             raise VDAuthError("Need to Re-Auth")
-        
+
         if api_response.status_code == 204:  # this means empty
             resp_json = {}
         else:
-            resp_json = api_response.json
+            try: 
+                resp_json = api_response.json
+            except:
+                resp_json = {"error": "error parsing json response"} 
 
         if isinstance(resp_json, list):
-            #wrap it in a multi container
+            # wrap it in a multi container
             return self.__RESPONSES_OBJECT__(self, resp_json, path_params,
                                              query_params, self.__RESPONSE_OBJECT__,
-                                             is_ok)
+                                             is_ok, payload, self.account_id)
 
         return self.__RESPONSE_OBJECT__(self, resp_json, path_params,
-                                        query_params,is_ok)
-    
+                                        query_params, is_ok, payload,
+                                        self.account_id)
+
+    @raw_response_retry
     def get_raw(self, path_param=None, reauth=False, **query_params):
         """
         Get the raw http response for this object.  This is rarely used by a
@@ -379,13 +451,13 @@ class _VDAPIService(object):
         """
         params = _format_params(query_params)
         return API(reauth=reauth).get(_format_url(self.endpoint, path_param),
-                                      params = params)
+                                      params=params)
 
     def get(self, path_param=None, reauth=False, **query_params):
         """
         Make a get request to this api service.  Allows you to pass in arbitrary
-        query paramaters. 
-        
+        query paramaters.
+
         Examples::
 
             # get all supply_tags
@@ -409,103 +481,145 @@ class _VDAPIService(object):
         try:
             return self.build_response(
                 self.get_raw(path_param, reauth=reauth, **query_params),
-                path_param, 
+                path_param,
                 query_params
             )
         except VDAuthError as e:
-            #we only retry if we are redo'n on an auto reauth 
+            # we only retry if we are redo'n on an auto reauth
             if not reauth:
                 _msg.info("Reauthing and then retry")
-                return self.get(path_param, reauth=True, **query_params) 
+                return self.get(path_param, reauth=True, **query_params)
             raise e
 
-    
+    @raw_response_retry
+    def _put_raw(self, path_param, data, reauth=False, **query_params):
+        params = _format_params(query_params)
+        return API(reauth=reauth).put(
+                    _format_url(self.endpoint, path_param),
+                    params=params,
+                    data=_json.dumps(data)
+                    )
+
     def put(self, path_param, data, reauth=False, **query_params):
         global API
 
         try:
-            params = _format_params(query_params)
             return self.build_response(
-                    API(reauth=reauth).put(
-                                          _format_url(self.endpoint, path_param), 
-                                           params = params,
-                                          data = _json.dumps(data)
-                             ),
-                    path_param, 
-                    query_params
+                    self._put_raw(path_param, data, reauth=reauth, **query_params),
+                    path_param,
+                    query_params,
+                    payload=data
             )
         except VDAuthError as e:
-            #we only retry if we are redo'n on an auto reauth 
+            # we only retry if we are redo'n on an auto reauth
             if not reauth:
                 _msg.info("Reauthing and then retry")
-                return self.put(path_param, data, reauth=True, **query_params) 
+                return self.put(path_param, data, reauth=True, **query_params)
             raise e
 
+    @raw_response_retry
+    def _post_raw(self, data, path_param="", reauth=False, files=None, **query_params):
+        params = _format_params(query_params)
 
-    def post(self, data, path_param = "", reauth=False, **query_params):
+        if not files:
+            return API(reauth=reauth).post(
+                    _format_url(self.endpoint, path_param),
+                    params=params,
+                    data=_json.dumps(data)
+                    )
+
+        m = MultipartEncoder(
+            fields=files
+        )
+        return API(reauth=reauth).post(
+                _format_url(self.endpoint, path_param),
+                headers={'Content-Type': m.content_type},
+                params=params,
+                data=m
+                )
+
+    def post(self, data, path_param="", files=None, reauth=False, **query_params):
+        global API
+        try:
+            return self.build_response(
+                self._post_raw(data, path_param, reauth=reauth, files=files, **query_params),
+                path_param,
+                query_params,
+                payload=data
+            )
+        except VDAuthError as e:
+            # we only retry if we are redo'n on an auto reauth
+            if not reauth:
+                _msg.info("Reauthing and then retry")
+                return self.post(data, path_param, reauth=True, files=files, **query_params)
+            # means that we had already tried a reauth and it failed
+            raise e
+
+    def delete(self, path_param="", reauth=False, **query_params):
         global API
         try:
             params = _format_params(query_params)
             return self.build_response(
-                    API(reauth=reauth).post(
-                                        _format_url(self.endpoint, path_param),
-                                        params=params,
-                                        data = _json.dumps(data)
-                             ),
-                    path_param, 
+                    API(reauth=reauth).delete(
+                        _format_url(self.endpoint, path_param),
+                        params=params,
+                        ),
+                    path_param,
                     query_params
             )
         except VDAuthError as e:
-            #we only retry if we are redo'n on an auto reauth 
+            # we only retry if we are redo'n on an auto reauth
             if not reauth:
                 _msg.info("Reauthing and then retry")
-                return self.new(data, path_param, reauth=True, **query_params) 
-            #means that we had already tried a reauth and it failed
+                return self.delete(path_param, reauth=True, **query_params)
+            # means that we had already tried a reauth and it failed
             raise e
 
-    def delete(self, path_param = "", reauth=False, **query_params):
-        global API
-        try:
-            params = _format_params(query_params)
-            return self.build_response(
-                    API(reauth=reauth).delete(_format_url(self.endpoint, path_param),
-                                              params = params,
-                             ),
-                    path_param, 
-                    query_params
+    def _raw_bulk_delete(self, data, path_param="", reauth=False, files=None, **query_params):
+        params = _format_params(query_params)
+        
+        if not files:
+            return API(reauth=reauth).delete(
+                _format_url(self.endpoint, path_param),
+                params=params,
+                data=_json.dumps(data)
+                )
+
+        m = MultipartEncoder(
+            fields=files
+        )
+
+        return API(reauth=reauth).delete(
+            _format_url(self.endpoint, path_param),
+            params=params,
+            headers={'Content-Type': m.content_type},
+            data=m
             )
-        except VDAuthError as e:
-            #we only retry if we are redo'n on an auto reauth 
-            if not reauth:
-                _msg.info("Reauthing and then retry")
-                return self.delete(path_param, reauth=True, **query_params) 
-            #means that we had already tried a reauth and it failed
-            raise e
 
-    def bulk_delete(self, data, path_param = "", reauth=False, **query_params):
+    def bulk_delete(self, data, path_param="", reauth=False, files=None, **query_params):
         """
-        Delete an object.  
+        Delete an object.
         """
         global API
         try:
-            params = _format_params(query_params)
             return self.build_response(
-                    API(reauth=reauth).delete(_format_url(self.endpoint, path_param),
-                                              params = params,
-                                              data = _json.dumps(data)
-                             ),
-                    path_param, 
+                self._raw_bulk_delete(data, path_param=path_param,
+                                      reauth=reauth, files=files,
+                                      **query_params),
+                    
+                    path_param,
                     query_params
             )
         except VDAuthError as e:
-            #we only retry if we are redo'n on an auto reauth 
+            # we only retry if we are redo'n on an auto reauth
             if not reauth:
                 _msg.info("Reauthing and then retry")
-                return self.delete(data, path_param, reauth=True, **query_params) 
-            #means that we had already tried a reauth and it failed
+                return self.buck_delete(data, path_param, reauth=True,
+                                        files=files, **query_params)
+            # means that we had already tried a reauth and it failed
             raise e
 
-    def new(self, data, path_param = "", reauth=False, **query_params):
+    def new(self, data, path_param="", reauth=False, **query_params):
         """
         Create a new object.  You need to pass in the required fields as a
         dictionary.  For instance::
@@ -514,37 +628,68 @@ class _VDAPIService(object):
             print resp.ok
         """
         return self.post(data, path_param, reauth, **query_params)
-       
- 
 
-from ._supply import _SupplyTagAPI, _SupplyPartnerAPI, _SupplyGroupAPI
-from ._demand import _DemandTagAPI, _DemandPartnerAPI, _DemandGroupAPI
-from ._common import _DomainListAPI, _BillAPI
+
+from ._supply import _SupplyTagAPI, _SupplyPartnerAPI, _SupplyLabelAPI, _ConnectedSupplyAPI, _SupplyRouterAPI
+from ._demand import _DemandTagAPI, _DemandPartnerAPI, _DemandLabelAPI, _ConnectedDemandAPI, _CampaignAPI, _SpotxConnectAPI
+from ._demand import _CreativeAPI, _TagCreativeAPI, _VideoCreativeAPI, _AudioCreativeAPI
+from ._common import _DomainListAPI, _BillAPI, _KeyAPI, _AppBundleListAPI, _AppNameListAPI, _IpListAPI, _SegmentListAPI, _AdvertiserDomainListAPI
+from ._common import _ChannelIdListAPI, _DealIdListAPI, _PlacementIdListAPI, _PublisherIdListAPI
 from ._reporting import _ReportingAPI, _TrafficQualityReport
 from ._account import _AccountAPI, _UserAPI
+from ._direct_connect import _DirectConnectionAPI
+from ._object_change_messages import _ObjectChangeMessagesAPI
 
 accounts = _AccountAPI()
+app_bundles = _AppBundleListAPI()
+app_names = _AppNameListAPI()
+advertiser_domain_lists = _AdvertiserDomainListAPI()
+ip_lists = _IpListAPI()
 
 bills = _BillAPI()
 
-demand_groups = _DemandGroupAPI()
+campaigns = _CampaignAPI()
+connected_demand = _ConnectedDemandAPI()
+connected_supply = _ConnectedSupplyAPI()
+channel_id_lists = _ChannelIdListAPI() 
+
+deal_id_lists = _DealIdListAPI() 
+demand_labels = _DemandLabelAPI()
 demand_tags = _DemandTagAPI()
 demand_partners = _DemandPartnerAPI()
+direct_connections = _DirectConnectionAPI()
+
 domain_lists = _DomainListAPI()
 
+creatives = _CreativeAPI()
+tag_creatives = _TagCreativeAPI()
+video_creatives = _VideoCreativeAPI()
+audio_creatives = _AudioCreativeAPI()
+
+keys = _KeyAPI()
+
+_object_change_messages = _ObjectChangeMessagesAPI()
+
+placement_id_lists = _PlacementIdListAPI()
+publisher_id_lists = _PublisherIdListAPI()
+
 reports = _ReportingAPI()
 quality_reports = _TrafficQualityReport()
 
-supply_groups = _SupplyGroupAPI()
+spotx_connects = _SpotxConnectAPI()
+supply_labels = _SupplyLabelAPI()
 supply_tags = _SupplyTagAPI()
+supply_routers = _SupplyRouterAPI()
 supply_partners = _SupplyPartnerAPI()
+segment_lists = _SegmentListAPI()
 
 users = _UserAPI()
 
+
 def raw_get(path_param, **query_params):
     global API
     params = _format_params(query_params)
-    return API().get(_format_url("", path_param), params = params).json
+    return API().get(_format_url("", path_param), params=params).json
 
 
 def _install_ipython_completers():  # pragma: no cover
@@ -557,17 +702,16 @@ def _install_ipython_completers():  # pragma: no cover
         Add in all the methods of the _wrapped object so its
         visible in iPython as well
         """
-        prev_completions+=obj._tab_completions()
+        prev_completions += obj._tab_completions()
         return prev_completions
 
 
 # Importing IPython brings in about 200 modules, so we want to avoid it unless
 # we're in IPython (when those modules are loaded anyway).
-# Code attributed to Pandas, Thanks Wes 
+# Code attributed to Pandas, Thanks Wes
 if "IPython" in _sys.modules:  # pragma: no cover
     try:
         _install_ipython_completers()
     except Exception:
         _msg.debug("Error loading tab completers")
-        pass 
-
+        pass
